/*
Presto - An Allegro tilemap library
Copyright (C) 2008  Peter Fernandes

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


presto.c: engine variables and functions
*/

#include "presto.h"

unsigned long int frame_delay_gcd;
unsigned int update_anim;

char tilemap_errstr[256+25+64];
char tilemap_fileerr[256+25];
int tilemap_err;



char *tilemap_get_err(void) {
	char *(strarray[5])={"End of tilemap file reached: ","Could not allocate memory",
	"Error reading from tilemap file, bad format probably: ","Could not load file: ",
	"Could not install Allegro timer"};
	
	
	if(tilemap_err<=-1 && tilemap_err>=-5) {
		strcpy(tilemap_errstr,"[Presto] ");
		strcpy(tilemap_errstr+9,strarray[-(tilemap_err+1)]);
		
		if(tilemap_err==-4 || tilemap_err==-3 || tilemap_err==-1) {
			strcpy(tilemap_errstr+9+(tilemap_err==-4?21:(tilemap_err==-3?54:29)),tilemap_fileerr);
		}
			
	}
	
	
	return tilemap_errstr;
}


void render_tilemap(BITMAP *dest, TILEMAP *tilemap, unsigned int x, unsigned int y, unsigned int width, unsigned int height) {
	unsigned int w=0;
	unsigned int h=0;
	unsigned int offset_w=x/tilemap->tile_w;
	unsigned int offset_h=y/tilemap->tile_h;
	TILE *map=tilemap->map+(offset_h*tilemap->width)+offset_w;

	set_alpha_blender();
	
	for(h=0;h<height;h++) {
		for(w=0;w<width;w++,map++) {
			
			if(tilemap->trans[map->tileid]>=0) {
				draw_trans_sprite(dest,tilemap->pool[map->tileid],(w*tilemap->tile_w)-(x%tilemap->tile_w),(h*tilemap->tile_h)-(y%tilemap->tile_h));
			}
			else {
				draw_sprite(dest,tilemap->pool[map->tileid],(w*tilemap->tile_w)-(x%tilemap->tile_w),(h*tilemap->tile_h)-(y%tilemap->tile_h));
			}
		}
		
		map=tilemap->map+((offset_h+h+1)*tilemap->width)+offset_w;
	}

	
}

/*reads tilemap and sprite list data from a file*/
int load_tilemap(char *filename, TILEMAP *tilemap) {
	FILE *in;
	int temp;
	unsigned int i;
	unsigned int f;
	unsigned int g;
	unsigned int h;
	int x;
	int y;
	TILE *pos;
	char buf;
	char name[257];
	
	
	if(!(in=fopen(filename,"rb"))) {
		tilemap_err=-4;
		strcpy(tilemap_fileerr,filename);
		return -4;
	}
	
	temp=fgetc(in);
	fseek(in,-1,SEEK_CUR);
	
	if(temp=='\n' || temp=='#') {
		
		if(next_line(in)) {
			temp=-1;
			goto error_close_file;
		}
	}
	
	
	
	//READ: tilemap width/height
	if(fscanf(in,"%d,%d",&tilemap->width,&tilemap->height)<=0) {
		temp=-3;
		goto error_close_file;
	}
	
	if(next_line(in)) {
		temp=-1;
		goto error_close_file;
	}
	
	//READ: tile width/height
	if(fscanf(in,"%d,%d",&tilemap->tile_w,&tilemap->tile_h)<=0) {
		temp=-3;
		goto error_close_file;
	}
	
	
	if(tilemap_create_dims(tilemap,tilemap->height,tilemap->width)) {
		temp=-2;
		goto error_close_file;
	}
	
	
	if(next_line(in)) {
		temp=-1;
		goto error_free_map;
	}
	
	
	//READ: poolsize
	if(fscanf(in,"%d",&tilemap->poolsize)<=0) {
		temp=-3;
		goto error_free_map;
	}
	
	if(next_line(in)) {
		temp=-1;
		goto error_free_map;
	}
	
	
	if(!(tilemap->pool=malloc(tilemap->poolsize*sizeof(BITMAP *)))) {
		temp=-2;
		goto error_free_map;
	}
	
	if(!(tilemap->trans=malloc(tilemap->poolsize*sizeof(int)))) {
		temp=-2;
		goto error_free_pool;
	}
	
	drawing_mode(DRAW_MODE_TRANS,NULL,0,0);
	set_write_alpha_blender();
	
	set_color_conversion(COLORCONV_KEEP_ALPHA);
	for(i=0;i<tilemap->poolsize;i++) {
		//READ: bitmap filename
		if(fscanf(in,"%256s %d",name,&tilemap->trans[i])<=0) {
			temp=-3;
			goto error_free_trans;
		}
		
		if(next_line(in)) {
			temp=-1;
			goto error_free_trans;
		}
		
		
		
		if(!strcmp(name+(strlen(name)-4),".tga") || !strcmp(name+(strlen(name)-4),".icb")) {
			if((tilemap->pool[i]=load_tga(name,NULL))==NULL) {
				temp=-4;
				strcpy(tilemap_fileerr,name);
				goto error_free_trans;
			}
			
			if(tilemap->trans[i]>=0) {
				for (y=0;y<tilemap->pool[i]->h;y++) {
					for (x=0;x<tilemap->pool[i]->w;x++) {
						if(geta(getpixel(tilemap->pool[i],x,y)))
							putpixel(tilemap->pool[i],x,y,tilemap->trans[i]);
					}
				}
			}
		}
		else if(!strcmp(name+(strlen(name)-4),".bmp")) {	
			if((tilemap->pool[i]=load_bitmap(name,NULL))==NULL) {
				temp=-4;
				strcpy(tilemap_fileerr,name);
				goto error_free_trans;
			}
			
			
			if(tilemap->trans[i]>=0)
				temp=tilemap->trans[i];
			else
				temp=255;
			
			for (y=0;y<tilemap->pool[i]->h;y++) {
				for (x=0;x<tilemap->pool[i]->w;x++) {
					putpixel(tilemap->pool[i],x,y,temp);
				}
			}
			
		}
		else if(!strcmp(name+(strlen(name)-4),".png")) {
			if((tilemap->pool[i]=load_png(name,NULL))==NULL) {
				fprintf(stderr,"HORJHIOWR\n");
				temp=-4;
				strcpy(tilemap_fileerr,name);
				goto error_free_trans;
			}
			
			if(tilemap->trans[i]>=0) {
				for (y=0;y<tilemap->pool[i]->h;y++) {
					for (x=0;x<tilemap->pool[i]->w;x++) {
						if(geta(getpixel(tilemap->pool[i],x,y)))
							putpixel(tilemap->pool[i],x,y,tilemap->trans[i]);
					}
				}
			}
		}
		else {
			temp=-4;
			strcpy(tilemap_fileerr,"Unsupported file format: ");
			strcpy(tilemap_fileerr+25,name);
			goto error_free_trans;
		}
		
	}
	
	
	
	//READ: number of animation orders to read
	if(fscanf(in,"%d",&f)<=0) {
		temp=-3;
		goto error_free_trans;
	}
	
	if(next_line(in)) {
		temp=-1;
		goto error_free_trans;
	}
	
	tilemap->num_anims=f;
	
	
	if(!(tilemap->orders=malloc(f*sizeof(unsigned int *)))) {
		temp=-2;
		goto error_free_trans;
	}
	
	if(!(tilemap->delays=malloc(f*sizeof(unsigned int *)))) {
		temp=-2;
		goto error_free_orders;
	}
	
	if(!(tilemap->loopvals=malloc(f*sizeof(int)))) {
		temp=-2;
		goto error_free_delays;
	}
	
	if(!(tilemap->num_frames=malloc(f*sizeof(int)))) {
		temp=-2;
		goto error_free_loopvals;
	}
	
	if(!(tilemap->default_frame=malloc(f*sizeof(int)))) {
		temp=-2;
		goto error_free_num_frames;
	}
	
	if(!(tilemap->frameclock=malloc(f*sizeof(unsigned int)))) {
		temp=-2;
		goto error_free_default_frame;
	}
	
	if(!(tilemap->wait=malloc(f*sizeof(unsigned int)))) {
		temp=-2;
		goto error_free_frameclock;
	}
	
	
	//initialize frameclock and global wait array
	memset(tilemap->frameclock,0,f*sizeof(unsigned int));
	
	for(i=0;i<f;i++) {
		tilemap->wait[i]=1;
	}
	
	
	
	//get each animation order
	for(i=0;i<f;i++) {
		
		//READ: number of frames to read / number of times to loop
		if(fscanf(in,"%d %d %d",&g,&tilemap->loopvals[i],&tilemap->default_frame[i])<=0) {
			temp=-3;
			h=0;
			g=0;
			goto error_free_inloop;
		}

		tilemap->num_frames[i]=g;

		
		if(!(tilemap->orders[i]=malloc(g*sizeof(unsigned int)))) {
			temp=-2;
			h=0;
			goto error_free_inloop;
		}
		
		if(!(tilemap->delays[i]=malloc(g*sizeof(unsigned int)))) {
			temp=-2;
			h=1;
			goto error_free_inloop;
		}
		
		for(h=0;h<g;h++) {
			//READ: a sprite index and delay time
			if(fscanf(in,"%d,%d",&tilemap->orders[i][h],&tilemap->delays[i][h])<=0) {
				temp=-3;
				goto error_free_orders;
			}
		
		}
		
		if(next_line(in)) {
			temp=-1;
			goto error_free_map;
		}
		
	}
	
	//DO NOT REUSE the variable f
	
	
	
	pos=tilemap->map;
	
	//now start reading tilemap data
	for(i=0;i<(tilemap->width*tilemap->height);i++,pos++) {
		
		//cue to the next number entry
		do {
			buf=fgetc(in);
		} while ((buf<'0' || buf>'9') && buf!='A' && buf!='S' && buf!=EOF);
		
		
		
		//it's an animated tile
		if(buf=='A' || buf=='S') {
			//READ: animation index
			if(fscanf(in,"%u",&temp)<=0) {
				temp=-3;
				goto error_free_map;
			}
			
			pos->orderid=temp;
			pos->tileid=tilemap->orders[temp][0];
			pos->index=0;
			pos->wait=1;
			pos->loop=tilemap->loopvals[temp];
			pos->sync_to_frameclock=(buf=='S');
			
			
		}
		else { //it's a static tile
			
			fseek(in,-1,SEEK_CUR);
			
			//READ: tile index
			if(fscanf(in,"%u",&temp)<=0) {
				temp=-3;
				goto error_free_map;
			}
			
			pos->tileid=temp;
			pos->orderid=-1;
		}

		
		
	}
	
	
	fclose(in);
	return 0;
	
	
	//_______CLEANUP SECTION________//
	
	error_free_frameclock:
	free(tilemap->frameclock);
	
	error_free_default_frame:
	free(tilemap->default_frame);
	
	error_free_num_frames:
	free(tilemap->num_frames);
	
	h=0;
	
	error_free_inloop:
	
	free(tilemap->wait);

	for(i=0;(i+1)<=(f+h);i++) {
		free(tilemap->orders[h]);
	}
	
	for(i=0;(i+1)<=f;i++) {
		free(tilemap->delays[h]);
	}
	
	error_free_loopvals:
	free(tilemap->loopvals);
	
	error_free_delays:
	free(tilemap->delays);
	
	error_free_orders:
	free(tilemap->orders);
	
	error_free_trans:
	free(tilemap->trans);
	
	error_free_pool:
	free(tilemap->pool);
	
	error_free_map:
	free(tilemap->map);
	
	error_close_file:
	fclose(in);
	
	if(temp==-1 || temp==-3) {
		strcpy(tilemap_fileerr,filename);
	}
	
	tilemap_err=temp;
	return temp;
}



void free_tilemap(TILEMAP *tilemap) {
	unsigned int i;
	
	
	for(i=0;i<tilemap->num_anims;i++) {
		free(tilemap->orders[i]);
		free(tilemap->delays[i]);
	}
	
	for(i=0;i<tilemap->poolsize;i++) {
		destroy_bitmap(tilemap->pool[i]);
	}
	
	
	free(tilemap->wait);
	free(tilemap->frameclock);
	free(tilemap->orders);
	free(tilemap->delays);
	free(tilemap->loopvals);
	free(tilemap->map);
	free(tilemap->pool);
}





//animation timer callback
void anim_timer(void) {
	
	update_anim=1;
	
}
END_OF_FUNCTION(anim_timer);



//Sets up the engine for animated tiles.  Does not have to be called if only using static tiles.
int install_tilemap(TILEMAP_LIST *list) {
	unsigned int i;
	unsigned int f;
	unsigned int h;
	
	frame_delay_gcd=0;
	update_anim=0;

	
	
	//calculate the LCM of all the delays for animated tiles
	for(f=0;f<list->num;f++) {
		for(i=0;i<list->list[f]->num_anims;i++) {
			for(h=0;h<list->list[f]->num_frames[i];h++) {
				frame_delay_gcd=GCD(list->list[f]->delays[i][h],frame_delay_gcd);
			}
			
		}
		
	}
	
	LOCK_VARIABLE(frame_count);
	LOCK_FUNCTION(anim_timer);
	
	
	if(install_int(anim_timer,frame_delay_gcd)<0) {
		tilemap_err=-5;
		return -5;
	}
	
	
	
	return 0;
}

void uninstall_tilemap(void) {

	remove_int(anim_timer);
}

//some useful math functions used for tile animation

unsigned int GCD(unsigned int a, unsigned int b) {
	unsigned int i=(a>b?b:a);
	unsigned int j=(a>b?a:b);
	
	if(a==0)
		return b;
	if(b==0)
		return a;
	
	for(;i;i--) {
		if(!(j%i))
			return i;
		
	}
	
	return 1;
}


void update_anim_tiles(TILEMAP_LIST *list, unsigned int x, unsigned int y, unsigned int width, unsigned int height) {
	unsigned int waits_needed;
	unsigned int f;
	unsigned int h;
	unsigned int w;
	TILEMAP *tilemap;
	
	
	
	if(update_anim) {
		for(f=0;f<list->num;f++) {
			tilemap=list->list[f];
			
			//update frameclock
			for(f=0;f<tilemap->num_anims;f++) {
				waits_needed=tilemap->delays[f][tilemap->frameclock[f]]/frame_delay_gcd;
				if(tilemap->wait[f]>=waits_needed) {
					tilemap->wait[f]=1;
					
					if(tilemap->frameclock[f]>=(tilemap->num_frames[f]-1)) {
						tilemap->frameclock[f]=0;				
					}
					else
						tilemap->frameclock[f]++;
					
				}
				else {
					tilemap->wait[f]++;
				}
				
				
			}
			
			
			for(h=y;h<(y+height) && h<tilemap->height;h++) {
				for(w=x;w<(x+width) && w<tilemap->width;w++) {
					//if this is an animated tile
					if(map(w,h).orderid>=0) {
						waits_needed=tilemap->delays[map(w,h).orderid][map(w,h).index]/frame_delay_gcd;
						
						if(map(w,h).loop) {
							if(map(w,h).sync_to_frameclock) {
								map(w,h).tileid=tilemap->orders[map(w,h).orderid][tilemap->frameclock[map(w,h).orderid]];
								map(w,h).index=tilemap->frameclock[map(w,h).orderid];
								
								if((!tilemap->frameclock[map(w,h).orderid]) && (tilemap->wait[map(w,h).orderid]==1)) {
									if(map(w,h).loop>0) {
										map(w,h).loop--;
										
										if((map(w,h).loop==0) && (tilemap->default_frame[map(w,h).orderid]>=0))
											map(w,h).tileid=tilemap->default_frame[map(w,h).orderid];

									}
								}
								
							}
							else {
								if(map(w,h).wait>=waits_needed) {
									map(w,h).wait=1;
									
									
									if(map(w,h).index>=(tilemap->num_frames[map(w,h).orderid]-1)) {
										map(w,h).index=0;
										
										if(map(w,h).loop>0)
											map(w,h).loop--;					
									}
									else
										map(w,h).index++;
										
									map(w,h).tileid=tilemap->orders[map(w,h).orderid][map(w,h).index];
								}
								else
									map(w,h).wait++;
							}
						}
					}
				}
			}
				
		}
		
		update_anim=0;
	}
}


//Seeks to the next line of real data in a tilemap file
int next_line(FILE *file) {
	char temp;
	
	while(1) {
		
		//go to start of the next line
		while((temp=getc(file))!='\n') {
			if(temp==-1)
				return -1;
		}

		empty_line:
		
		temp=fgetc(file);
		
		if(temp=='\n')
			goto empty_line;
		if(temp=='#')
			continue;
		if(temp==-1)
			return -1;

		
		break;
	}
	
	fseek(file,-1,SEEK_CUR);
	
	return 0;
}



